/*
 *
 * Diff to HTML (diff2html.js)
 * Author: rtfpessoa
 *
 */

global.diffStatementNewLines = 0;
global.coveredDiffStatementNewLines = 0;
global.uncoveredDiffStatementNewLines = 0;
global.nonExecutableDiffStatementNewLines = 0;
global.fileCoverageData = {};
global.files = [];
global.coverage = null;
global.coveredLines = 0;
global.uncoveredLines = 0;
global.overallStatementCoverageRate = 0;
global.fileDisplayMap = {};
global.displayMode = 'standard';

(function() {
  var diffParser = require('./diff-parser.js').DiffParser;
  var htmlPrinter = require('./html-printer.js').HtmlPrinter;
  var utils = require('./utils.js').Utils;

  function Diff2Html() {
  }

  var defaultConfig = {
    inputFormat: 'diff',
    outputFormat: 'line-by-line',
    showFiles: false,
    matching: 'none',
    matchWordsThreshold: 0.25,
    matchingMaxComparisons: 2500,
    maxLineSizeInBlockForComparison: 200,
    maxLineLengthHighlight: 10000,
    templates: {},
    rawTemplates: {},
    renderNothingWhenEmpty: false
  };

  /*
   * Generates json object from string diff input
   */
  Diff2Html.prototype.getJsonFromDiff = function(diffInput, config) {
    var cfg = utils.safeConfig(config, defaultConfig);
    return diffParser.generateDiffJson(diffInput, cfg);
  };

  /*
   * Generates the html diff. The config parameter configures the output/input formats and other options
   */
  Diff2Html.prototype.getPrettyHtml = function(diffInput, config, coverage, coverageFilePatterns, workspace, customFileNameGeneratorFile, displayMode, strict, caseInsensitive) {
    var cfg = utils.safeConfig(config, defaultConfig);

    var diffJson = diffInput;
    if (!cfg.inputFormat || cfg.inputFormat === 'diff') {
      diffJson = diffParser.generateDiffJson(diffInput, cfg);
    }

    var diffOutput = '';
    if (cfg.outputFormat === 'side-by-side') {
      diffOutput = htmlPrinter.generateSideBySideJsonHtml(diffJson, cfg);
    } else {
      global.coverage = coverage;
      global.displayMode = displayMode;
      diffOutput = htmlPrinter.generateLineByLineJsonHtml(diffJson, cfg, coverage, coverageFilePatterns, workspace, customFileNameGeneratorFile, displayMode, strict, caseInsensitive);
    }

    var fileList = '';
    if (cfg.showFiles === true) {
      fileList = htmlPrinter.generateFileListSummary(diffJson, cfg);
    }

    return this.getMetadata(fileList, diffOutput);
  };

  Diff2Html.prototype.getMetadata = function(fileList, diffOutput) {
    var effectiveChangedLines = global.coveredDiffStatementNewLines + global.uncoveredDiffStatementNewLines;
    var diffStatementCoverageRate = effectiveChangedLines === 0 ? 100 : parseFloat((global.coveredDiffStatementNewLines / effectiveChangedLines * 100).toFixed(2));
    for (var fileName in global.fileCoverageData) {
      var covered = global.fileCoverageData[fileName].covered;
      var uncovered = global.fileCoverageData[fileName].uncovered;
      var total = covered + uncovered;
      if (total === 0) {
        global.fileCoverageData[fileName].coverageRate = null;
      } else {
        global.fileCoverageData[fileName].coverageRate = parseFloat((covered / total * 100).toFixed(2));
      }
    }
    const fileCoverageData = [];
    for (const fileName in global.fileCoverageData) {
      const data = global.fileCoverageData[fileName];
      fileCoverageData.push({
        fileName,
        covered: data.covered,
        uncovered: data.uncovered,
        coverageRate: data.coverageRate
      });
    }
    fileCoverageData.sort((a, b) => {
      return a.coverageRate - b.coverageRate;
    });
    var metadata = {
      diffStatementNewLines: global.diffStatementNewLines,
      coveredDiffStatementNewLines: global.coveredDiffStatementNewLines,
      uncoveredDiffStatementNewLines: global.uncoveredDiffStatementNewLines,
      nonExecutableDiffStatementNewLines: global.nonExecutableDiffStatementNewLines,
      diffStatementCoverageRate,
      overallStatementCoverageRate: global.overallStatementCoverageRate,
      fileCoverageData,
      files: global.files
    };
    return {
      fileList,
      htmlContents: diffOutput,
      metadata
    };
  };

  /*
   * Deprecated methods - The following methods exist only to maintain compatibility with previous versions
   */

  /*
   * Generates pretty html from string diff input
   */
  Diff2Html.prototype.getPrettyHtmlFromDiff = function(diffInput, config) {
    var cfg = utils.safeConfig(config, defaultConfig);
    cfg.inputFormat = 'diff';
    cfg.outputFormat = 'line-by-line';
    return this.getPrettyHtml(diffInput, cfg);
  };

  /*
   * Generates pretty html from a json object
   */
  Diff2Html.prototype.getPrettyHtmlFromJson = function(diffJson, config) {
    var cfg = utils.safeConfig(config, defaultConfig);
    cfg.inputFormat = 'json';
    cfg.outputFormat = 'line-by-line';
    return this.getPrettyHtml(diffJson, cfg);
  };

  /*
   * Generates pretty side by side html from string diff input
   */
  Diff2Html.prototype.getPrettySideBySideHtmlFromDiff = function(diffInput, config) {
    var cfg = utils.safeConfig(config, defaultConfig);
    cfg.inputFormat = 'diff';
    cfg.outputFormat = 'side-by-side';
    return this.getPrettyHtml(diffInput, cfg);
  };

  /*
   * Generates pretty side by side html from a json object
   */
  Diff2Html.prototype.getPrettySideBySideHtmlFromJson = function(diffJson, config) {
    var cfg = utils.safeConfig(config, defaultConfig);
    cfg.inputFormat = 'json';
    cfg.outputFormat = 'side-by-side';
    return this.getPrettyHtml(diffJson, cfg);
  };

  var diffObject = new Diff2Html();
  module.exports.Diff2Html = diffObject;

  // Expose diff2html in the browser
  global.Diff2Html = diffObject;
})();
