/*
 *
 * LineByLinePrinter (line-by-line-printer.js)
 * Author: rtfpessoa
 *
 */

(function() {
  var minimatch = require('minimatch');
  var path = require('path');
  var diffParser = require('./diff-parser.js').DiffParser;
  var printerUtils = require('./printer-utils.js').PrinterUtils;
  var utils = require('./utils.js').Utils;
  var Rematch = require('./rematch.js').Rematch;

  var hoganUtils;

  var genericTemplatesPath = 'generic';
  var baseTemplatesPath = 'line-by-line';
  var iconsBaseTemplatesPath = 'icon';
  var tagsBaseTemplatesPath = 'tag';

  function LineByLinePrinter(config) {
    this.config = config;

    var HoganJsUtils = require('./hoganjs-utils.js').HoganJsUtils;
    hoganUtils = new HoganJsUtils(config);
  }

  LineByLinePrinter.prototype.makeFileDiffHtml = function(file, diffs) {
    if (this.config.renderNothingWhenEmpty && file.blocks && !file.blocks.length) return '';

    var fileDiffTemplate = hoganUtils.template(baseTemplatesPath, 'file-diff');
    var filePathTemplate = hoganUtils.template(genericTemplatesPath, 'file-path');
    var fileIconTemplate = hoganUtils.template(iconsBaseTemplatesPath, 'file');
    var fileTagTemplate = hoganUtils.template(tagsBaseTemplatesPath, printerUtils.getFileTypeIcon(file));

    return fileDiffTemplate.render({
      file: file,
      fileHtmlId: printerUtils.getHtmlId(file),
      diffs: diffs,
      filePath: filePathTemplate.render({
        fileDiffName: printerUtils.getDiffName(file)
      }, {
        fileIcon: fileIconTemplate,
        fileTag: fileTagTemplate
      })
    });
  };

  LineByLinePrinter.prototype.makeLineByLineHtmlWrapper = function(content) {
    return hoganUtils.render(genericTemplatesPath, 'wrapper', {'content': content});
  };

  LineByLinePrinter.prototype.generateLineByLineJsonHtml = function(diffFiles, coverage, coverageFilePatterns, workspace, customFileNameGeneratorFile, displayMode, strict, caseInsensitive) {
    var that = this;
    that.coverageFilePatterns = coverageFilePatterns;
    that.workspace = workspace;
    var htmlDiffs = diffFiles.map(function(file) {
      var diffs;
      global.fileDisplayMap[file.newName] = true;
      if (file.blocks.length) {
        diffs = that._generateFileHtml(file, coverage, customFileNameGeneratorFile, displayMode, strict, caseInsensitive);
      } else {
        diffs = that._generateEmptyDiff();
      }
      if (diffs === '') {
        return '';
      }
      return that.makeFileDiffHtml(file, diffs);
    });

    const maxLen = Math.pow(2, 28);
    const blocks = [];
    let block = [];
    let len = 0;
    htmlDiffs.forEach(line => {
      if (len + line.length > maxLen) {
        blocks.push(block);
        block = [];
        len = 0;
      }
      len += line.length;
      block.push(line);
    });
    if (block.length > 0) {
      blocks.push(block);
    }
    return blocks.map(block => this.makeLineByLineHtmlWrapper(block.join('\n')));
  };

  var matcher = Rematch.rematch(function(a, b) {
    var amod = a.content.substr(1);
    var bmod = b.content.substr(1);

    return Rematch.distance(amod, bmod);
  });

  LineByLinePrinter.prototype.makeColumnLineNumberHtml = function(block) {
    return hoganUtils.render(genericTemplatesPath, 'column-line-number', {
      diffParser: diffParser,
      blockHeader: utils.escape(block.header),
      lineClass: 'd2h-code-linenumber',
      contentClass: 'd2h-code-line'
    });
  };

  LineByLinePrinter.prototype._generateFileHtml = function(file, coverage, customFileNameGeneratorFile, displayMode, strict, caseInsensitive) {
    var that = this;
    const fileName = this._getFileName(file.newName, coverage, customFileNameGeneratorFile, caseInsensitive);
    global.files.push(fileName);
    const isExcludeFile = this._isExcludeFile(file.newName) || !this._isExecutableFile(file.newName);
    let fileHasUncoveredLines = false;

    function getNewLineCoverageStatus(file, newLine, coverage) {
      // line hits: 0: uncovered, >0: covered, -1: non-executable
      if (isExcludeFile) {
        newLine.hits = -1;
        that._increaseDiffStatementNewLines(fileName, newLine.hits);
        return;
      }
      if (!coverage) {
        newLine.hits = 0;
        fileHasUncoveredLines = true;
        that._increaseDiffStatementNewLines(fileName, newLine.hits);
        return;
      }
      newLine.hits = -1;
      const fileCoverage = coverage[fileName];
      if (!fileCoverage) {
        if (!strict) {
          newLine.hits = 0; // mark line uncovered if coverage data not found for current file in non-strict mode
          fileHasUncoveredLines = true;
        } else {
          newLine.hits = -1; // mark line non-executable if coverage data not found for current file in strict mode
          fileHasUncoveredLines = false;
        }
        that._increaseDiffStatementNewLines(fileName, newLine.hits);
        return;
      }
      const newLineNewNumber = newLine.newNumber;
      const lineMap = fileCoverage.lineMap;
      const line = lineMap[newLineNewNumber];
      if (!line) {
        newLine.hits = -1;
      } else {
        newLine.hits = line.hits;
        if (line.hits === 0) {
          fileHasUncoveredLines = true;
        }
      }
      that._increaseDiffStatementNewLines(fileName, newLine.hits);
    }

    const results = file.blocks.map(function(block) {
      var lines = that.makeColumnLineNumberHtml(block);
      var oldLines = [];
      var newLines = [];

      function processChangeBlock() {
        var matches;
        var insertType;
        var deleteType;

        var comparisons = oldLines.length * newLines.length;

        var maxLineSizeInBlock = Math.max.apply(null,
          [0].concat((oldLines.concat(newLines)).map(
            function(elem) {
              return elem.content.length;
            }
          )));

        var doMatching = comparisons < that.config.matchingMaxComparisons &&
          maxLineSizeInBlock < that.config.maxLineSizeInBlockForComparison &&
          (that.config.matching === 'lines' || that.config.matching === 'words');

        if (doMatching) {
          matches = matcher(oldLines, newLines);
          insertType = diffParser.LINE_TYPE.INSERT_CHANGES;
          deleteType = diffParser.LINE_TYPE.DELETE_CHANGES;
        } else {
          matches = [[oldLines, newLines]];
          insertType = diffParser.LINE_TYPE.INSERTS;
          deleteType = diffParser.LINE_TYPE.DELETES;
        }

        matches.forEach(function(match) {
          oldLines = match[0];
          newLines = match[1];

          var processedOldLines = [];
          var processedNewLines = [];

          var common = Math.min(oldLines.length, newLines.length);

          var oldLine, newLine;
          for (var j = 0; j < common; j++) {
            oldLine = oldLines[j];
            newLine = newLines[j];

            that.config.isCombined = file.isCombined;
            var diff = printerUtils.diffHighlight(oldLine.content, newLine.content, that.config);

            processedOldLines +=
              that.makeLineHtml(file.isCombined, deleteType, oldLine.oldNumber, oldLine.newNumber,
                diff.first.line, diff.first.prefix, oldLine.hits);

            processedNewLines +=
              that.makeLineHtml(file.isCombined, insertType, newLine.oldNumber, newLine.newNumber,
                diff.second.line, diff.second.prefix, newLine.hits);
          }

          lines += processedOldLines + processedNewLines;
          lines += that._processLines(file.isCombined, oldLines.slice(common), newLines.slice(common));
        });

        oldLines = [];
        newLines = [];
      }

      for (var i = 0; i < block.lines.length; i++) {
        var line = block.lines[i];
        var escapedLine = utils.escape(line.content);

        if (line.type !== diffParser.LINE_TYPE.INSERTS &&
          (newLines.length > 0 || (line.type !== diffParser.LINE_TYPE.DELETES && oldLines.length > 0))) {
          processChangeBlock();
        }

        if (line.type === diffParser.LINE_TYPE.CONTEXT) {
          lines += that.makeLineHtml(file.isCombined, line.type, line.oldNumber, line.newNumber, escapedLine, '', line.hits);
        } else if (line.type === diffParser.LINE_TYPE.INSERTS && !oldLines.length) {
          getNewLineCoverageStatus(file, line, coverage);
          lines += that.makeLineHtml(file.isCombined, line.type, line.oldNumber, line.newNumber, escapedLine, '', line.hits);
        } else if (line.type === diffParser.LINE_TYPE.DELETES) {
          oldLines.push(line);
        } else if (line.type === diffParser.LINE_TYPE.INSERTS && Boolean(oldLines.length)) {
          getNewLineCoverageStatus(file, line, coverage);
          newLines.push(line);
        } else {
          console.error('Unknown state in html line-by-line generator');
          processChangeBlock();
        }
      }

      processChangeBlock();

      return lines;
    }).join('\n');

    if (displayMode === 'simplify' && (isExcludeFile || !fileHasUncoveredLines)) {
      global.fileDisplayMap[file.newName] = false;
      return '';
    }
    return results;
  };

  LineByLinePrinter.prototype._processLines = function(isCombined, oldLines, newLines) {
    var lines = '';

    for (var i = 0; i < oldLines.length; i++) {
      var oldLine = oldLines[i];
      var oldEscapedLine = utils.escape(oldLine.content);
      lines += this.makeLineHtml(isCombined, oldLine.type, oldLine.oldNumber, oldLine.newNumber, oldEscapedLine);
    }

    for (var j = 0; j < newLines.length; j++) {
      var newLine = newLines[j];
      var newEscapedLine = utils.escape(newLine.content);
      lines += this.makeLineHtml(isCombined, newLine.type, newLine.oldNumber, newLine.newNumber, newEscapedLine, '', newLine.hits);
    }

    return lines;
  };

  LineByLinePrinter.prototype.makeLineHtml = function(isCombined, type, oldNumber, newNumber, content, possiblePrefix, newLineHits) {
    let newNumberContent = utils.valueOrEmpty(newNumber);
    var lineNumberTemplate = hoganUtils.render(baseTemplatesPath, 'numbers', {
      oldNumber: utils.valueOrEmpty(oldNumber),
      newNumber: newNumberContent
    });

    var lineWithoutPrefix = content;
    var prefix = possiblePrefix;

    if (!prefix) {
      var lineWithPrefix = printerUtils.separatePrefix(isCombined, content);
      prefix = lineWithPrefix.prefix;
      lineWithoutPrefix = lineWithPrefix.line;
    }

    if (prefix === ' ') {
      prefix = '&nbsp;';
    }

    var lineHitClass = 'line-hits-off';
    var hits = newLineHits > 0;
    var noHits = newLineHits === 0;
    var hitsOff = newLineHits === -1;
    var isDel = type === 'd2h-del';

    if (newLineHits === -1) {
      lineHitClass = 'line-hits-off';
    } else if (newLineHits === 0) {
      lineHitClass = 'line-no-hits';
    } else if (newLineHits > 0) {
      lineHitClass = 'line-hits';
    }

    var lineHitContainerClass = isDel ? 'd2h-code-line-hits-del' : 'd2h-code-line-hits-ins';

    return hoganUtils.render(genericTemplatesPath, 'line',
      {
        type: type,
        lineClass: 'd2h-code-linenumber',
        contentClass: 'd2h-code-line',
        lineHitContainerClass,
        lineHitClass,
        prefix: prefix,
        content: lineWithoutPrefix,
        lineNumber: lineNumberTemplate,
        newLineHits,
        hits,
        noHits,
        hitsOff,
        isDel
      });
  };

  LineByLinePrinter.prototype._generateEmptyDiff = function() {
    return hoganUtils.render(genericTemplatesPath, 'empty-diff', {
      contentClass: 'd2h-code-line',
      diffParser: diffParser
    });
  };

  LineByLinePrinter.prototype._increaseDiffStatementNewLines = function(fileName, lineHits) {
    global.diffStatementNewLines += 1;
    if (lineHits === 0) {
      global.uncoveredDiffStatementNewLines += 1;
    } else if (lineHits > 0) {
      global.coveredDiffStatementNewLines += 1;
    } else if (lineHits === -1) {
      global.nonExecutableDiffStatementNewLines += 1;
    }
    this._collectFileCoverageData(fileName, lineHits);
  };

  LineByLinePrinter.prototype._collectFileCoverageData = function(fileName, lineHits) {
    if (lineHits === -1) { // don't collect coverage data of non-executable files
      return;
    }
    if (!global.fileCoverageData[fileName]) {
      global.fileCoverageData[fileName] = {
        covered: 0,
        uncovered: 0
      };
    }
    if (lineHits === 0) {
      global.fileCoverageData[fileName].uncovered += 1;
    } else if (lineHits > 0) {
      global.fileCoverageData[fileName].covered += 1;
    }
  };

  LineByLinePrinter.prototype._isExcludeFile = function(fileName) {
    if (!this.coverageFilePatterns) {
      return false;
    }
    var exclude = true;
    for (var i = 0; i < this.coverageFilePatterns.length; i += 1) {
      var group = this.coverageFilePatterns[i];
      for (var j = 0; j < group.length; j += 1) {
        if (!minimatch(fileName, group[j])) {
          break;
        }
        if (j === group.length - 1) {
          exclude = false;
        }
      }
      if (!exclude) {
        break;
      }
    }
    return exclude;
  };

  LineByLinePrinter.prototype._isExecutableFile = function(fileName) {
    const extensionMap = {
      '.js': true,
      '.ts': true,
      '.jsx': true,
      '.tsx': true,
      '.cpp': true,
      '.c': true,
      '.java': true,
      '.kt': true,
      '.swift': true,
      '.cs': true,
      '.m': true,
      '.mm': true
    };
    var extName = path.extname(fileName);
    return !!extensionMap[extName];
  };

  LineByLinePrinter.prototype._getFileName = function(originalFileName, coverage, customFileNameGeneratorFile, caseInsensitive) {
    if (customFileNameGeneratorFile) {
      const filePath = utils.resolveFilePath(customFileNameGeneratorFile);
      const FileNameGenerator = require(filePath);
      const fileNameGenerator = new FileNameGenerator();
      const possibleFileNames = fileNameGenerator.generateFileNames(originalFileName);
      for (const possibleFileName of possibleFileNames) {
        const filename = utils.getCoverageFilePath(possibleFileName, caseInsensitive);
        if (coverage[filename]) {
          return filename;
        }
      }
    } else {
      return utils.getCoverageFilePath(path.normalize(path.join(this.workspace || '', originalFileName)), caseInsensitive);
    }
    return utils.getCoverageFilePath(originalFileName, caseInsensitive);
  };

  module.exports.LineByLinePrinter = LineByLinePrinter;
})();
